/***** ms.c     ************************************************
*
*       Generates samples of gametes ( theta given or fixed number 
*						of segregating sites.)
*	Usage is shown by typing ms without arguments.   
        usage: ms nsam howmany  -t  theta  [options]
		or
	       ms nsam howmany -s segsites  [options] 

	   nsam is the number of gametes per sample.
	   howmany is the number of samples to produce.
	   With -t the numbers of segregating sites will randomly vary 
		from one sample to the next.
	   with -s segsites,  the number of segregating sites will be
		segsites in each sample.

           Other options:

              -r rho nsites     (recombination)
              -c f track_len    (gene conversion:  To use this option,
				 the -r option must also be used, and rho must be > 0. )   
              -m mig_rate npop n1 n2 ...  (Island model with symmetric migration.)
              -d  nintn  nr1 np1 t1  [nr2 np2 t2 ... ]  (population size changes) 

*	  Arguments of the options are explained here:

           npop:  Number of subpopulations which make up the total population
           ni:  the sample size from the i th subpopulation (all must be 
		specified.) The output will have the gametes in order such that
		the first n1 gametes are from the first island, the next n2 are
		from the second island, etc.
           nsites: number of sites between which recombination can occur.
           theta: 4No times the neutral mutation rate 
           rho: recombination rate between ends of segment times 4No
	   f: ratio of conversion rate to recombination rate. (Wiuf and Hein model.)
	   track_len:  mean length of conversion track in units of sites.  The 
		       total number of sites is nsites, specified with the -r option.
           mig_rate: migration rate: the fraction of each subpop made up of
                 migrants times 4No. 
           howmany: howmany samples to generate.
	   nintn:  number of intervals with specified demographic parameters.
		(Following nintn there must be 3*nintn arguments specifying the
		values of the following three parameters for each interval.
		starting with the most recent interval and proceeding into the past.
		The population growth rate implied by the values of the last
		set of three values will apply into the indefinite past.)
	   nri:  the population size (relative to No) at the beginning (most 
		recent point) of the i-th  interval.
	   npi:  the population size (relative to No) at the end (farthest in 
		the past point) of the i-th interval.
	   ti:  The duration (in units of 4No) of the i-th interval.

	Note:  In the above definition, No is the total diploid population if
		npop is one, otherwise, No is the diploid population size of each
		subpopulation. However, if the -d option is used the population sizes
		are No times various factors specified by the -d arguments.
	A seed file called "seedms" will be created  if it doesn't exist. The
		seed(s) in this file will be modified by the program. 
		So subsequent runs
		will produce new output.  The initial contents of seedms will be
		printed on the second line of the output.
        Output consists of one line with the command line arguments and one
	 	line with the seed(s).
		The samples appear sequentially following that line.
		Each sample begins with "//", then the number of segregating sites, the positions
		of the segregating sites (on a scale of 0.0 - 1.0). On the following
		lines are the sampled gametes, with mutants alleles represented as
		ones and ancestral alleles as zeros.
	To compile:  cc -o ms  ms.c  streec.c  rand1.c -lm
		or:  cc -o ms ms.c streec.c rand2.c -lm
	 (Of course, gcc would be used instead of cc on some machines.  And -O3 or 
		some other optimization switches might be usefully employed with some 
		compilers.) ( rand1.c uses drand48(), whereas rand2.c uses rand() ).

example runs:  ms 7 3 -t 2.0 -r 1.0 1000 -d 1 100. 1. .75  -m .5 2 4 3 >sample3.out
 	   with no seedms file present   
	produces the results shown in the file sample3.out .
	These are samples from a two-island (symmetric migration) with 4Nom = 0.5.
	Sample size 4 from one subpopulation, sample size 3 from the other.  There 
	1000 sites between which recombination can occur.  rho  = 1.0.  4Nou (the 
	mutation parameter) = 2.0.  The population exponentially shrinks as one
	goes back in time, such that the population size .75*4No generations ago
	it is No, while the present population size is 100No. The exponential
	 shrinking continues in this case indefinitely into the past. 

	An additional example of the use of -d:
       ms 10 1000 -t 2.0 -d 3 .5 .5 1.0  2.0 2.0 2.0  1.0 1.0 1.0 
	will produce 1000 samples of size 10, for a demographic history in which
	 the population size is
	No/2 for 4No generations, then is  2No for 2*4No generations, then is No
	indefinitely into the past beyond that. 
*
*   Modifications made to combine ms and mss on 25 Feb 2001
*	Modifications to command line options to use switches  25 Feb 2001
*	Modifications to add // before each sample  25 Feb 2001
	Modifications to add gene conversion 5 Mar 2001
	Added demographic options -d  13 Mar 2001
	Changed ran1() to use rand(). Changed seed i/o to accomodate this change. 20 April.
	Changed cleftr() to check for zero rand() .13 June 2001
	Move seed stuff to subroutine seedit()  11 July 2001
***************************************************************************/



#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define SITESINC 10 
#define REFLOCIMAX 1000
#define MAX1st 5.0
#define MAX2nd 2.0

unsigned maxsites = SITESINC ;

struct node{
	int abv;
	int ndes;
	float time;
	};

struct segl {
	int beg;
	struct node *ptree;
	int next;
	};

double *posit ;


main(argc,argv)
        int argc;
        char *argv[];
{
        int nsam ,nsites, i, ns, howmany, npop, *config, pop, count, narg; 
     double  theta, r,  mig_rate, estimator, gst, alpha, within, between;
        char **list, **cmatrix();
        FILE *pf, *fopen() ;
        double ttot,  f, track_len ;
        int  segsites, segsitesin, nintn  ;
	double *nrec, *npast, *tpast ;
	int getpars(int argc, char *argv[], int *, int *, double *, double *, double *, int *,
     	  double *, int *,  int *, int **, double *,
	  int *, double **, double **, double ** );
	void seedit( const char * ) ;
        
        /* added paras and functions */
        int x, y, z, tmp1;
        int S, Su;
        float K, Ksum, Hsum, a1, Tdiv, Ttmp, tmp2, tmpave, tmpsum1, tmpsum2, Sdivsum, TdivMAX, TdivMAXcheck, Ktmp;
        int Sdiv, locus, locusn, accept, Sdivtmp;
        
		char buf[256];
        int data2, data3;    
        float data1, data4, data5;    
		int nsamOBS[REFLOCIMAX+3];
		float LOBS[REFLOCIMAX+3], KOBS[REFLOCIMAX+3], DOBS[REFLOCIMAX+3], LOBSsum, KOBSsum, DOBSsum, rOBS, thetapsOBS, Kcheck[REFLOCIMAX+3], Sdivcheck[REFLOCIMAX+3], tmphet;
		float rhothetaratio;
        FILE *outfile, *infile;
        if ((outfile = fopen( "div.infile", "w")) == NULL)
	{
		printf("Could not open output file\n");
		exit(-1);
	}
        if ((infile = fopen( "refloci", "r")) == NULL)
	{
		printf("Could not open input file\n");
		exit(-1);
	}

printf ("\nList of Reference Loci **********************************\n");
locusn=0;
        if (infile != NULL)
	{
            while (fgets(buf, 256, infile) != NULL)
            {
				locusn++;
				sscanf(buf, "%*s\t%f\t%d\t%d\t%f\t%f\t%f\n", &data1, &data2, &data3, &data4, &data5);
				printf ("L=%f, n=%d, S=%d, K=%f, Div=%f\n", data1, data2, data3, data4, data5);
				LOBS[locusn]=data1;
				nsamOBS[locusn]=data2;
				KOBS[locusn]=data4;
				DOBS[locusn]=data5;
				
                
            }
	}
	
LOBSsum=KOBSsum=DOBSsum=0;
for (x=1; x<=locusn; x++)
{
	LOBSsum += LOBS[x];
	KOBSsum += KOBS[x];
	DOBSsum += DOBS[x];
}
rOBS=KOBSsum/DOBSsum;
thetapsOBS=KOBSsum/LOBSsum;
printf ("*********************************************************\n%d regions: LOBSsum=%f, KOBSsum=%f, DOBSsum=%f, rOBS=%f, theta/site=%f\n\n\n",  locusn, LOBSsum, KOBSsum, DOBSsum, rOBS, thetapsOBS);
        
        Ksum=Sdivsum=0;
        TdivMAX = MAX1st/rOBS;
        
        /* end added paras and functions */
        
        
       
	 getpars( argc, argv,  &nsam, &howmany, &r, &f, &track_len, &nsites,
     	   &theta, &segsitesin,  &npop, &config, &mig_rate,
	   &nintn, &nrec, &npast, &tpast );
	for( i=0; i<argc; i++) printf("%s ",argv[i]);
        for( i=0; i<argc; i++) fprintf(outfile, "%s ",argv[i]); fprintf(outfile, "\n");
        fprintf(outfile, "%d\n", howmany);
    
        
	

	seedit( "s");
        pf = stdout ;

       if( theta > 0.0 ) {
	   list = cmatrix(nsam,maxsites+1);
	   posit = (double *)malloc( (unsigned)( maxsites*sizeof( double)) ) ;
	   }
	else {
	    list = cmatrix(nsam, segsitesin+1 ) ;
	    posit = (double *)malloc( (unsigned)( segsitesin*sizeof( double)) ) ;
	   }




        count=0;
        accept=0;
		rhothetaratio=r/theta;
    
    
        for (locus=1; locus<=locusn; locus++)  Kcheck[locus]=Sdivcheck[locus]=0;
    for (locus=1; locus<=locusn; locus++)  if (nsam<nsamOBS[locus])  {printf ("ERROR: input the maximum sample size of the reference regions\n\n");  exit(-1);}
    
    
	printf("TdivMAX=%f\n", TdivMAX);  TdivMAXcheck=0;
	/*while( howmany-count++ ) */
    while( howmany-accept ) 
    {
       Tdiv=drand48()*TdivMAX;
       Sdiv=K=0;
       count++;
            
       for (locus=1; locus<=locusn; locus++)
       {

			
			theta=LOBS[locus]*thetapsOBS;
			nsites=LOBS[locus];
			r=theta*rhothetaratio;
			/*printf ("count=%d, locus=%d, nsam=%d (but use %d), theta=%f, r=%f, nsites=%d --> ", count, locus, nsam, nsamOBS[locus], theta, r, nsites);*/


			
			
			segsites = gensam(npop,nsam,config,nsites, theta , segsitesin, r, f,track_len, mig_rate,list,nintn, nrec, npast, tpast );
       
            Ktmp=0;
			if( segsites > 0 ) 
            {
                /* checcking theta */
                for(x=0; x<segsites; x++)  
                {
                    tmp1=0;
                    for(z=0; z<nsamOBS[locus]; z++)  if (list[z][x]=='1')  tmp1++;
                    if (tmp1>0 && tmp1<nsamOBS[locus])
                    {
                        tmphet=2.0*tmp1*(nsamOBS[locus]-tmp1)/nsamOBS[locus]/(nsamOBS[locus]-1);
						Ktmp+=tmphet;
					
						
						
                    }
					
					/*for(z=0; z<nsamOBS[locus]; z++)  printf ("%c", list[z][x]);  printf ("  %f(%f)\n", tmphet, K);*/
                }
            }
            
			K+=Ktmp;
			
            Ttmp=Tdiv-log(1.-drand48());
            Sdivtmp=poisso(Ttmp*theta);
			Sdiv+=Sdivtmp;
			
			Kcheck[locus]+=Ktmp;
			Sdivcheck[locus]+=Sdivtmp;
			
			/*printf ("K=%f(%f), Div=%d(%d)\n", Ktmp, K, Sdivtmp, Sdiv);*/

            
            
       }
       Ksum+=K;
       Sdivsum+=Sdiv;

       if (K/Sdiv>(0.99*rOBS) && K/Sdiv<1.01*rOBS)  
       {
            accept ++;
			if (TdivMAXcheck<Tdiv)  TdivMAXcheck=Tdiv;   
			if (accept==100)  TdivMAX=TdivMAXcheck*MAX2nd;
            fprintf (outfile ,"%f\n", Tdiv);
            /*printf ("accept %d/%d --- Tdiv=%f, theta=%f, r=%f, n=%d --- K=%f, Sdiv=%d, r=%f -- %f/%f\n", accept, count, Tdiv, thetapsOBS, thetapsOBS*rhothetaratio, nsam, K, Sdiv, K/Sdiv, TdivMAXcheck, TdivMAX);*/
        }
    
       if (count%10000==0) printf ("%d --- Tdiv=%f, theta=%f, r=%f, n=%d --- K=%f, Sdiv=%d --- accept = %d\n", count, Tdiv, thetapsOBS,  thetapsOBS*rhothetaratio, nsam, K, Sdiv, accept);
       
       
       
    }
    
    
    printf("\n");
    
    
    /*for (locus=1; locus<=locusn; locus++)
	{
		a1=0;  for (x=1; x<nsamOBS[locus]; x++)  a1 += 1.0/x;
		theta=LOBS[locus]*thetapsOBS;
		nsites=LOBS[locus];
		r=theta*rhothetaratio;
		printf ("locus=%2d, n=%2d, theta=%2.2f, r=%2.2f, nsites=%5.2d -->\t", locus, nsamOBS[locus], theta, r, nsites);
		printf("Observation: K = %3.2f | %3.2f\t\t", Kcheck[locus]/count, theta);
		printf("Observation: Sdiv = %3.2f | %3.2f\n", 1.0*Sdivcheck[locus]/count, theta*(TdivMAX/2+1));
	}*/
    
              
     
    seedit( "end" );
}








	int 
gensam( npop,nsam,inconfig, nsites, theta, segsites, r, f, track_len,mig_rate,list,
	nintn, nrec, npast, tpast  )
	int npop,nsam,inconfig[], nsites, segsites, nintn ; 
	char **list;
	double theta, r, mig_rate, f, track_len, *nrec, *npast, *tpast;
{
	int nsegs, h, i, k, j, seg, ns, start, end, len, segsit ;
	struct segl *seglst, *segtre_mig() ; /* used to be: [MAXSEG];  */
	double nsinv,  tseg, tt, ttime() ;
	double *pk;
	int *ss;


	nsinv = 1./nsites;
	seglst = segtre_mig(npop,nsam,inconfig,nsites,r,f,track_len,mig_rate, &nsegs, nintn, nrec, npast, tpast) ;
	
    if( theta > 0.0 ) {
	ns = 0 ;
	for( seg=0, k=0; k<nsegs; seg=seglst[seg].next, k++) { 
		end = ( k<nsegs-1 ? seglst[seglst[seg].next].beg -1 : nsites-1 );
		start = seglst[seg].beg ;
		len = end - start + 1 ;
		tseg = len*(theta/nsites) ;
		tt = ttime(seglst[seg].ptree, nsam);
		segsit = poisso( tseg*tt );
		if( (segsit + ns) >= maxsites ) {
			maxsites = segsit + ns + SITESINC ;
			posit = (double *)realloc(posit, maxsites*sizeof(double) ) ;
			  biggerlist(nsam, list) ; 
			}
		make_gametes(nsam,seglst[seg].ptree,tt, segsit, ns, list );
		free(seglst[seg].ptree) ;
		locate(segsit,start*nsinv, len*nsinv,posit+ns);   
		ns += segsit;
		}
    }
   else {

        pk = (double *)malloc((unsigned)(nsegs*sizeof(double)));
        ss = (int *)malloc((unsigned)(nsegs*sizeof(int)));
        if( (pk==NULL) || (ss==NULL) ) perror("malloc error. gensam.2");


	tt = 0.0 ;
	for( seg=0, k=0; k<nsegs; seg=seglst[seg].next, k++) { 
		end = ( k<nsegs-1 ? seglst[seglst[seg].next].beg -1 : nsites-1 );
		start = seglst[seg].beg ;
		len = end - start + 1 ;
		tseg = len/(double)nsites ;
               pk[k] = ttime(seglst[seg].ptree,nsam)*tseg ;
                tt += pk[k] ;
		}
       for (k=0;k<nsegs;k++) pk[k] /= tt ;
        mnmial(segsites,nsegs,pk,ss);
	ns = 0 ;
	for( seg=0, k=0; k<nsegs; seg=seglst[seg].next, k++) { 
		end = ( k<nsegs-1 ? seglst[seglst[seg].next].beg -1 : nsites-1 );
		start = seglst[seg].beg ;
		len = end - start + 1 ;
                tseg = len/(double)nsites;
                make_gametes(nsam,seglst[seg].ptree,tt*pk[k]/tseg, ss[k], ns, list);

		free(seglst[seg].ptree) ;
		locate(ss[k],start*nsinv, len*nsinv,posit+ns);   
		ns += ss[k] ;
		}
	free(pk);
	free(ss);

    }
	for(i=0;i<nsam;i++) list[i][ns] = '\0' ;
	return( ns ) ;
}

	int
biggerlist(nsam,  list )
	int nsam ;
	char ** list ;
{
	int i;

/*  fprintf(stderr,"maxsites: %d\n",maxsites);  */	
	for( i=0; i<nsam; i++){
	   list[i] = (char *)realloc( list[i],maxsites*sizeof(char) ) ;
	   if( list[i] == NULL ) perror( "realloc error. bigger");
	   }
}
	   


/* allocates space for gametes (character strings) */
	char **
cmatrix(nsam,len)
	int nsam, len;
{
	int i;
	char **m;

	if( ! ( m = (char **) malloc( (unsigned) nsam*sizeof( char* ) ) ) )
	   perror("alloc error in cmatrix") ;
	for( i=0; i<nsam; i++) {
		if( ! ( m[i] = (char *) malloc( (unsigned) len*sizeof( char ) )))
			perror("alloc error in cmatric. 2");
		}
	return( m );
}



	int
locate(n,beg,len,ptr)
	int n;
	double beg, len, *ptr;
{
	int i;

	ordran(n,ptr);
	for(i=0; i<n; i++)
		ptr[i] = beg + ptr[i]*len ;

}


int
getpars(int argc, char *argv[], int *pnsam, int *phowmany, double *pr, double *pf, double *ptrack_len, int *pnsites,
     double *ptheta, 
	int *pss, int *pnpop, int **pconfig, 
	double *pmig_rate, int *pnintn, double **pnrec, double **pnpast,
	double **ptpast 
	 )
{
	int arg, i, sum  ;
	
	if( argc < 4 ) usage();
	
	*pnsam = atoi( argv[1] );
	*phowmany = atoi( argv[2] );
	arg = 3 ;
	*pr = *ptheta =  *pmig_rate = *pf = 0.0 ;
	*ptrack_len = 0. ;
	*pnpop = 1 ;
	*pss = 0 ;
	*pconfig = (int *) malloc( (unsigned)(( *pnpop +1 ) *sizeof( int)) );
	(*pconfig)[0] = *pnsam ;
	*pnsites = 2 ;
	*pnintn = 1 ;
	*pnrec = (double *)malloc( (unsigned)( *pnintn+1)*sizeof( double) );
	*pnpast = (double *)malloc( (unsigned)( *pnintn+1)*sizeof( double) );
	*ptpast = (double *)malloc( (unsigned)( *pnintn+1)*sizeof( double) );
	*(*pnrec + 1 ) = 1.0 ;
	*(*pnpast + 1 ) = 1.0 ;
	*(*ptpast ) = 0.0 ;
	*(*ptpast + 1 ) = 1.0 ;

	while( arg < argc ){
		switch ( argv[arg][1] ){
			case 'r' : 
				arg++;
				*pr = atof(  argv[arg++] );
				*pnsites = atoi( argv[arg++]);
				if( *pnsites <2 ){
					fprintf(stderr,"with -r option must specify both rec_rate and nsites>1\n");
					usage();
					}
				break;		
			case 'c' : 
				arg++;
				*pf = atof(  argv[arg++] );
				*ptrack_len = atof( argv[arg++]);
				if( *ptrack_len <1. ){
					fprintf(stderr,"with -c option must specify both f and track_len>0\n");
					usage();
					}
				break;		
			case 't' : 
				arg++;
				*ptheta = atof(  argv[arg++] );
				break;
			case 's' : 
				arg++;
				*pss = atoi(  argv[arg++] );
				break;
			case 'm' : 
				arg++;
				*pmig_rate = atof(  argv[arg++] );
				*pnpop = atoi( argv[arg++]);
				*pconfig = (int *) realloc( *pconfig, (unsigned)(( *pnpop +1)*sizeof( int)));
				for( i=0; i< *pnpop; i++) (*pconfig)[i] = atoi( argv[arg++]);
				break;
			case 'd' : 
				arg++;
				*pnintn = atoi( argv[arg++] );
	*pnrec = (double *)realloc(*pnrec, (unsigned)( *pnintn+1)*sizeof( double) );
	*pnpast = (double *)realloc(*pnpast,(unsigned)( *pnintn+1)*sizeof( double) );
	*ptpast = (double *)realloc(*ptpast,(unsigned)( *pnintn+1)*sizeof( double) );
				for( i=1 ; i<= *pnintn ; i++){
				   *(*pnrec + i) = atof( argv[arg++] ) ;
				   *(*pnpast + i) = atof( argv[arg++] ) ;
				   *(*ptpast + i) = atof( argv[arg++] )+ *(*ptpast + i -1)  ;
				  }
				/*if( *(*ptpast + i)  <= 0.0 ) {
					fprintf(stderr," -d option error\n");
					usage();
					}*/
				break;
			default:  usage() ;
			}
		}
		if( (*ptheta == 0.0) && ( *pss == 0 ) ) {
			fprintf(stderr," either -s or -t option must be used. \n");
			usage();
			exit(1);
			}
		sum = 0 ;
		for( i=0; i< *pnpop; i++) sum += (*pconfig)[i] ;
		if( sum != *pnsam ) {
			fprintf(stderr," sum sample sizes != nsam\n");
			usage();
			exit(1);
			}
		return ;
}
	
	int
usage()
{
fprintf(stderr,"usage: ms nsam howmany \n");
fprintf(stderr,"  Options: \n"); 
fprintf(stderr,"\t -t theta   (this option or the next must be used.)\n");
fprintf(stderr,"\t -s segsites   ( fixed number of segregating sites)\n");
fprintf(stderr,"\t -r rho nsites     (rho here is 4Nc)\n");
fprintf(stderr,"\t -c f track_len   (ratio of conversion rate to rec rate) \n");
fprintf(stderr,"\t -m mig_rate npop n1 n2 ...   \n");    
fprintf(stderr,"\t -d  nintn nr1  np1 t1 [nr2 np2 t2 ...]   \n");
fprintf(stderr," See header to ms.c for explanation of these parameters.\n");

exit(1);
}
	
/************ make_gametes.c  *******************************************
*
*
*****************************************************************************/

#define STATE1 '1'
#define STATE2 '0'

	int
make_gametes(nsam, ptree,tt,newsites, ns, list )
	int nsam, newsites, ns;
	struct node *ptree;
	char **list;
	double tt ;
{
	int  tip, j,  node ;

	for(  j=ns; j< ns+newsites ;  j++ ) {
		node = pickb(  nsam, ptree, tt);
		for( tip=0; tip < nsam ; tip++) {
		   if( tdesn(ptree, tip, node) ) list[tip][j] = STATE1 ;
		   else list[tip][j] = STATE2 ;
		   }
		}
}


/***  ttime.c : Returns the total time in the tree, *ptree, with nsam tips. **/

	double
ttime( ptree, nsam)
	struct node *ptree;
	int nsam;
{
	double t;
	int i;

	t = (ptree + 2*nsam-2) -> time ;
	for( i=nsam; i< 2*nsam-1 ; i++)
		t += (ptree + i)-> time ;
	return(t);
}

/***  pickb : returns a random branch from the tree. The probability of picking
              a particular branch is proportional to its duration. tt is total
	      time in tree.   ****/

	int
pickb(nsam, ptree, tt)
	int nsam;
	struct node *ptree;
	double tt;
{
	double x, y, ran1();
	int i;

	x = ran1()*tt;
	for( i=0, y=0; i < 2*nsam-2 ; i++) {
		y += (ptree + (ptree+i)->abv )->time - (ptree+i)->time ;
		if( y >= x ) return( i ) ;
		}
	return( i );
}

/****  tdesn : returns 1 if tip is a descendant of node in *ptree, otherwise 0. **/

	int
tdesn(ptree, tip, node )
	struct node *ptree;
	int tip, node;
{
	int k;

	for( k= tip ; k < node ; k = (ptree+k)->abv ) ;
	if( k==node ) return(1);
	else return(0);
}


/* pick2()  */

	int
pick2(n,i,j)
	int n, *i, *j;
{
	double ran1();

	*i = n * ran1() ;
	while( ( *j = n * ran1() ) == *i )
		;
	return(0) ;
}

/**** ordran.c  ***/

	int
ordran(n,pbuf)
	int n;
	double pbuf[];
{
	ranvec(n,pbuf);
	order(n,pbuf);
	return;
}


	int
mnmial(n,nclass,p,rv)
	int n, nclass, rv[];
	double p[];
{
	double ran1();
	double x, s;
	int i, j;

	for(i=0; i<nclass; i++) rv[i]=0;
	for(i=0; i<n ; i++) {
	   x = ran1();
	   j=0;
	   s = p[0];
	   while( (x > s) && ( j<(nclass-1) ) )  s += p[++j];
	   rv[j]++;
	   }
	return(j);
}

        int
order(n,pbuf)
        int n;
        double pbuf[];
{
        int gap, i, j;
        double temp;

        for( gap= n/2; gap>0; gap /= 2)
           for( i=gap; i<n; i++)
                for( j=i-gap; j>=0 && pbuf[j]>pbuf[j+gap]; j -=gap) {
                   temp = pbuf[j];
                   pbuf[j] = pbuf[j+gap];
                   pbuf[j+gap] = temp;
                   }
}


	int
ranvec(n,pbuf)
	int n;
	double pbuf[];
{
	int i;
	double ran1();

	for(i=0; i<n; i++)
		pbuf[i] = ran1();

	return;
}



	int
poisso(u)
	double u;
{
	double  cump, ru, ran1(), p, gasdev() ;
	int i=1;

	if( u > 30. ) return( (int)(0.5 + gasdev(u,u)) );
	ru = ran1();
	p = exp(-u);
	if( ru < p) return(0);
	cump = p;
	
	while( ru > ( cump += (p *= u/i ) ) )
		i++;
	return(i);
}


/* a slight modification of crecipes version */

double gasdev(m,v)
	double m, v;
{
	static int iset=0;
	static float gset;
	float fac,r,v1,v2;
	double ran1();

	if  (iset == 0) {
		do {
			v1=2.0*ran1()-1.0;
			v2=2.0*ran1()-1.0;
			r=v1*v1+v2*v2;
		} while (r >= 1.0);
		fac=sqrt(-2.0*log(r)/r);
		gset= v1*fac;
		iset=1;
		return( m + sqrt(v)*v2*fac);
	} else {
		iset=0;
		return( m + sqrt(v)*gset ) ;
	}
}





